﻿//*****************************************************************************
// Description     : Main sample program to measure histogram
// Target Device   : APV8508, APV8104
// Target Platform : Windows, .NET Framework, C++
// Date            : August 2017
// Version         : 1.0
//*****************************************************************************
#include <Winsock2.h>
#include "file.h"
#include "sitcp.h"
#include "device.h"
#include "device_define.h"
#include "local_define.h"
#include "configHist.h"

using namespace System::Runtime::InteropServices;
using namespace System::Threading;


//
// Local functions declaration
//
static int config_device(SiTCPcom^, unsigned qword);
static int run_measurement(SiTCPcom^, unsigned qword);
static int receive_data(SiTCPcom^);


//#########################################################################
//  Name      : main
//  Function  : entry point
//  Arguments :
//      in args : parameters
//  Return :
//      0 : success
//      1 : failure
//  Note : 
//#########################################################################
int main(array<System::String ^> ^args)
{
    int run_ready;
    int ret;
    DEVICE_COM_INFO ^comInfo = gcnew DEVICE_COM_INFO;
    unsigned qword meas_count;

    run_ready = 0;

    // 1. connect to the device
    ret = device_connect(comInfo);

    if (ret != 0) {
        return(1);
    }

    // 2. set parameters to the device
    meas_count = device_msec2count(MEASUREMENT_TIME);   // convert msec to count
    ret = config_device(comInfo, meas_count);
    if (ret == 0) {
        run_ready = 1;
    } else {
        Console::WriteLine("setting device config error({0})...", ret);
    }

    if (run_ready == 1) {   // ready to run
        // 3. start measurement
        Console::WriteLine("start measurement!!!");
        ret = run_measurement(comInfo, meas_count);

        Console::WriteLine("finish measurement...");

        if (ret >= 0) {
            Console::WriteLine("measurement completed!!!");
        }
        else {
            Console::WriteLine("measurement error({0})...\n", ret);
        }
    }

    // 4. post processing
    device_disconnect(comInfo);

    return(0);
}


//#########################################################################
//  Name      : config_device
//  Function  : Set parameters to device
//  Arguments :
//      in comInf   : instance of DEVICE_COM_INFO class
//      in ms_count : measurement time (counts)
//  Return :
//       0 : success
//      -n : failed at nth parameter in configHist()
//      -10001: failed at mode setting
//      -10002: failed at measurement mode setting
//      -10003: failed at measurement time setting
//  Note : 
//#########################################################################
static int config_device(DEVICE_COM_INFO ^comInf, unsigned qword ms_count)
{
    int i, ret;

    // set by each command
    for (i = 0; configHist[i].cmd_addr != 0; i++) {
        ret = device_write_16bit(comInf, configHist[i].cmd_addr, configHist[i].value);
        if (ret < 0)  {
            return(-1 * (i+1));
        }
    }
    // wait for completion of each command processing
    Thread::Sleep(100);

    // set mode to histogram
    ret = device_write_16bit(comInf, DEVICE_COMMAND_MOD, DEVICE_PARAM_MOD_HIST);
    if (ret < 0)  {
        return(-10001);
    }

    // set measurement mode to real time
    ret = device_write_16bit(comInf, DEVICE_COMMAND_MMD, DEVICE_PARAM_MMD_REALTIME);
    if (ret < 0)  {
        return(-10002);
    }

    // set measurement time
    ret = device_write_64bit(comInf, DEVICE_COMMAND_MTM, ms_count);

    if (ret < 0) {
        return(-10003);
    }

    return(0);
}


//#########################################################################
//  Name      : run_measurement
//  Function  : Measurement processing
//  Arguments :
//      in comInf   : instance of DEVICE_COM_INFO class
//      in ms_count : measurement time (counts)
//  Return :
//       0 : success
//      -1 : failed while measurement
//      -2 : failed to write file header
//      -3 : failed to receive histogram data
//  Note : 
//#########################################################################
static int run_measurement(DEVICE_COM_INFO ^comInf, unsigned qword ms_count)
{
    int ret;
    unsigned qword now_count;
    unsigned int now_msec;
    unsigned short status;
    unsigned int output_count[DEVICE_MAX_CHNUM];
    unsigned int output_cps[DEVICE_MAX_CHNUM];

    // clear data
    device_write_3step(comInf, DEVICE_COMMAND_CLR);
    // wait for completion of clear processing
    Thread::Sleep(100);

    // start measurement
    device_start_measurement(comInf);
    // display top 2 lines of status field to explain each column
    Console::WriteLine("<time(msec)>   |<output rate(cps)>");
    if (DEVICE_MAX_CHNUM > 4) {
        Console::WriteLine(" real/total    |  CH1     CH2     CH3     CH4     CH5     CH6     CH7     CH8");
        Console::WriteLine("---------------+-------+-------+-------+-------+-------+-------+-------+-------");
    }else{
        Console::WriteLine(" real/total    |  CH1     CH2     CH3     CH4");
        Console::WriteLine("---------------+-------+-------+-------+-------");
    }

    while(1) {
        // read status
        ret = device_read_status(comInf, &status, &now_count, output_count, output_cps);

        if (ret < 0) { // communication error
            // quit measurement
            break;
        }
        // status is 'stopped' or measurement time elapsed
        if ((status == DEVICE_PARAM_AQS_STOP) || (now_count >= ms_count)) {
            break;
        }

        Thread::Sleep(200);   // wait for 200ms
    }

    // stop measurement
    device_stop_measurement(comInf);

    // quit by error
    if (ret < 0) {
        return(-1);
    }

    // write file header
    now_msec  = (unsigned int)(device_count2msec(now_count));    // convert count to msec
    ret =  file_write_header(SAVE_FILENAME, now_msec, output_count, output_cps);
    if (ret < 0) {
        return(-2);
    }

    // receive histogram data then write out to file
    ret = receive_data(comInf);
    if (ret < 0) {
        return(-3);
    } else {
        return(0);
    }
}


//#########################################################################
//  Name      : receive_data
//  Function  : receiving collected data
//  Arguments :
//      in comInf : instance of DEVICE_COM_INFO class
//  Return :
//       0 : success
//      -1 : failed to receive data
//      -2 : failed to write file
//  Note : 
//#########################################################################
static int receive_data(DEVICE_COM_INFO ^comInf)
{
    int ret, i;
    int result;
    unsigned int hdata[DEVICE_MAX_CHNUM][NUM_DATA];

    result = 0;

    // receive histogram data
    memset(hdata, 0, sizeof(hdata));
    for (i = 0; i < DEVICE_MAX_CHNUM; i++) {
        ret = device_read_histogram(comInf,
                                   (char*)(&hdata[i]), sizeof(hdata[i]),
                                    DEVICE_TIMEOUT_DATA, i);
    }

    if (ret > 0) {
        // write data to file
        ret =  file_write_data(SAVE_FILENAME, hdata);
        if (ret < 0) {
            // failed to write
            result = -2;
        }
    } else {
        // failed to receive data
        result = -1;
    }

    return(result);
}
