﻿//*****************************************************************************
// Description     : Definition of functions for SiTCP communication
// Target Device   : APV8508, APV8104
// Target Platform : Windows, .NET Framework, C++
// Date            : August 2017
// Version         : 1.0
//*****************************************************************************
#include <Winsock2.h>
#include "sitcp.h"
#include "command.h"
#include "device_define.h"
#include "local_define.h"

using namespace System::Runtime::InteropServices;
using namespace System::Threading;

//
// Local functions declaration
//
static int sitcp_read(UdpClient^, unsigned int, int, SITCP_FORMAT*);
static int sitcp_write(UdpClient^, unsigned int, void*, int);
static int sitcp_recv_cmd(UdpClient^, char*, int);


//#########################################################################
//  Name      : sitcp_connect
//  Function  : create UdpClient and TcpClient instance and connect to device
//  Arguments :
//      out comRef : SiTCPcom class instance
//  Return :
//       0 : success
//      -1 : failed to connect
//  Note 
//#########################################################################
int sitcp_connect(SiTCPcom ^comRef)
{
    // create new instance of UdpClient and TcpClient
    UdpClient ^udpClient = gcnew UdpClient(DEVICE_PORT_CTRL_OWNPC);
    TcpClient ^tcpClient = gcnew TcpClient;

    // change the size of the receive buffer
    udpClient->Client->ReceiveBufferSize = SITCP_SOCKBUF_CMD;
    tcpClient->Client->ReceiveBufferSize = SITCP_SOCKBUF_DATA;

    // connect to the device
    try {
        udpClient->Connect(IPADDR_DEVICE, DEVICE_PORT_CTRL_DEVICE);
        tcpClient->Connect(IPADDR_DEVICE, DEVICE_PORT_DATARECV);
    }
    catch (Exception ^e)
    {
        // connect error
        Console::WriteLine("device connect error ...");
        Console::WriteLine(e->Message);

        udpClient->Close();
        tcpClient->Close();

        return(-1);
    }

    // memorize instances for future reference
    comRef->cmdClient  = udpClient;
    comRef->dataClient = tcpClient;

    return(0);
}


//#########################################################################
//  Name      : sitcp_disconnect
//  Function  : disconnect device
//  Arguments :
//      in comRef : SiTCPcom class instance
//  Return : none
//  Note :
//#########################################################################
void sitcp_disconnect(SiTCPcom ^comRef)
{
    // close UDP connection and dispose UdpClient instance
    if ((comRef->cmdClient != nullptr) && (comRef->cmdClient->Client->Connected == true)) {
        (comRef->cmdClient)->Close();
    }

    // close TCP connection and dispose TcpClient instance
    if ((comRef->dataClient != nullptr) && (comRef->dataClient->Connected == true)) {
        (comRef->dataClient)->Close();
    }
}


//#########################################################################
//  Name      : sitcp_read
//  Function  : command communication(read request)
//  Arguments :
//      in  cmdClient : instance of UdpClient class for command communication
//      in  cmd_addr  : command address
//      in  size      : size of data to be read(byte)
//      out ackdt     : response data
//  Return :
//       0 : success
//      -1 : no response within a period of timeout value
//      -2 : invalid size of received packet
//      -3 : received packet is not a response to the request
//  Note : This function doesn't convert received data. It's up to caller.
//#########################################################################
int sitcp_read(UdpClient ^cmdClient, unsigned int cmd_addr, int size, SITCP_FORMAT *ackdt)
{
    SITCP_FORMAT reqdt, rcvdt;
    SITCP_HEADER cmpdt;
    int result;
    int rbyte;
    int ssize, rsize;
    unsigned char wid;
    static unsigned char gid = 0;   // global ID to distinguish response
    int i;
    unsigned char send_buff[sizeof(SITCP_FORMAT)], *pt;
    array<Byte> ^arr_byt = gcnew array< Byte >(sizeof(SITCP_FORMAT));

    result = 0;
    wid = gid++;

    // clear area for reception
    memset(&rcvdt, 0, sizeof(rcvdt));

    // set size of packet for request and response
    ssize = sizeof(SITCP_HEADER);
    rsize = sizeof(SITCP_HEADER) + size;

    // construct packet for read request
    memset(&reqdt, 0, sizeof(reqdt));
    reqdt.header.ver_type = SITCP_VER_TYPE;
    reqdt.header.cmd_flag = SITCP_READ_REQ_CMDFLAG;
    reqdt.header.id       = wid;
    reqdt.header.length   = (unsigned char)(size);
    reqdt.header.address  = (unsigned int)IPAddress::HostToNetworkOrder( (int)cmd_addr );

    // convert the packet to array of bytes
    pt = (unsigned char *)&reqdt;
    for (i = 0; i < sizeof(reqdt); i++, pt++) {
        send_buff[i] = *pt;
    }
    Marshal::Copy((IntPtr)send_buff, arr_byt, 0, sizeof(reqdt));

    // send the packet to the target
    cmdClient->Send(arr_byt, ssize);

    // receive response
    rbyte = sitcp_recv_cmd(cmdClient, (char*)(&rcvdt), DEVICE_TIMEOUT_CMD);

    if (rbyte >= rsize) {
        // construct header part of response packet from request one
        // by change flag, which is only difference, from 'request' to 'acknowledge'
        cmpdt = reqdt.header;
        cmpdt.cmd_flag = SITCP_READ_ACK_CMDFLAG;

        // check if header part is expected one
        if (memcmp(&cmpdt, &rcvdt.header, sizeof(cmpdt)) != 0) {
            result = -3;   // not a response to the request
        }

        // convert command(address) from network byte order to host byte order
        rcvdt.header.address = (unsigned int)IPAddress::NetworkToHostOrder( (int)(rcvdt.header.address) );
    }
    else {  // short of data size
        result = -2;
    }

    // pass response data to caller
    ackdt[0] = rcvdt;

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_16bit
//  Function  : command communication(read request for 16bit value)
//  Arguments :
//      in  comRef   : SiTCPcom class instance for referring UdpClient
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned short sitcp_read_16bit(SiTCPcom ^comRef, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    unsigned short result;

    result = 0;

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(comRef->cmdClient, cmd_addr, sizeof(result), &rdat);

    if (ret == 0) {
        result = (unsigned short)IPAddress::NetworkToHostOrder( (short)rdat.data.u16[0] );
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_32bit
//  Function  : command communication(read request for 32bit value)
//  Arguments :
//      in  comRef   : SiTCPcom class instance for referring UdpClient
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned int sitcp_read_32bit(SiTCPcom ^comRef, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    unsigned int result;

    result = 0;

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(comRef->cmdClient, cmd_addr, sizeof(result), &rdat);

    if (ret == 0) {
        result = (unsigned int)IPAddress::NetworkToHostOrder( (int)rdat.data.u32[0] );
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_48bit
//  Function  : command communication(read request for 48bit value)
//  Arguments :
//      in  comRef   : SiTCPcom class instance for referring UdpClient
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned qword sitcp_read_48bit(SiTCPcom ^comRef, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    char abyte[8];
    unsigned qword result;

    result = (unsigned qword)(0);

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(comRef->cmdClient, cmd_addr, 6, &rdat);

    if (ret == 0) {
        *(short*)(&abyte[0]) = 0x00;   // clear top 2bytes
        memcpy(&abyte[2], &rdat.data.u08[0], 6);
        result = (unsigned qword)IPAddress::NetworkToHostOrder( (qword) *(unsigned qword*)(&abyte[0]) );
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_read_64bit
//  Function  : command communication(read request for 64bit value)
//  Arguments :
//      in  comRef   : SiTCPcom class instance for referring UdpClient
//      in  cmd_addr : command address
//      out errcd    : error info; see return value of sitcp_read
//  Return : read value or 0(in case of error)
//  Note : This function converts received data from network byte order
//         to host byte order
//#########################################################################
unsigned qword sitcp_read_64bit(SiTCPcom ^comRef, unsigned int cmd_addr, int *errcd)
{
    SITCP_FORMAT rdat;
    int ret;
    unsigned qword result;

    result = (unsigned qword)(0);

    memset(&rdat, 0, sizeof(rdat));
    ret = sitcp_read(comRef->cmdClient, cmd_addr, sizeof(result), &rdat);

    if (ret == 0) {
        result = (unsigned qword)IPAddress::NetworkToHostOrder( (qword)rdat.data.u64[0] );
    }

    if (errcd != NULL) {
        *errcd = ret;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_write
//  Function  : command communication(write request)
//  Arguments :
//      in  cmdClient : instance of UdpClient class for command communication
//      in  cmd_addr  : command address
//      in  wbuffer   : data to be written 
//      in  size      : size of data to be written(byte)
//  Return :
//       0 : success
//      -2 : invalid size of received packet
//      -3 : received packet is not a response to the request
//  Note : This function doesn't convert written data. It's up to caller.
//#########################################################################
int sitcp_write(UdpClient ^cmdClient, unsigned int cmd_addr, void *wbuffer, int size)
{
    SITCP_FORMAT reqdt, rcvdt;
    SITCP_HEADER cmpdt;
    int result;
    int rbyte;
    int ssize, rsize;
    unsigned char wid;
    static unsigned char gid = 0;   // global ID to distinguish response
    int i;
    unsigned char send_buff[sizeof(SITCP_FORMAT)], *pt;
    array<Byte> ^arr_byt = gcnew array< Byte >(sizeof(SITCP_FORMAT));

    result = 0;
    wid = gid++;

    // clear area for reception
    memset(&rcvdt, 0, sizeof(rcvdt));

    // set size of packet for request and response
    ssize = sizeof(SITCP_HEADER) + size;
    rsize = ssize;

    // construct packet for write request
    memset(&reqdt, 0, sizeof(reqdt));
    reqdt.header.ver_type = SITCP_VER_TYPE;
    reqdt.header.cmd_flag = SITCP_WRITE_REQ_CMDFLAG;
    reqdt.header.id       = wid;
    reqdt.header.length   = (unsigned char)(size);
    reqdt.header.address  = (unsigned int)IPAddress::HostToNetworkOrder( (int)cmd_addr );
    memcpy(&reqdt.data, wbuffer, size);

    // convert the packet to array of bytes
    pt = (unsigned char *)&reqdt;
    for (i = 0; i < sizeof(reqdt); i++, pt++) {
        send_buff[i] = *pt;
    }
    Marshal::Copy((IntPtr)send_buff, arr_byt, 0, sizeof(reqdt));

    // send the packet to the target
    cmdClient->Send(arr_byt, ssize);

    // receive response
    rbyte = sitcp_recv_cmd(cmdClient, (char*)(&rcvdt), DEVICE_TIMEOUT_CMD);

    if (rbyte >= rsize) {
        // construct header part of response packet from request one
        // by change flag, which is only difference,  from 'request' to 'acknowledge'
        cmpdt = reqdt.header;
        cmpdt.cmd_flag = SITCP_WRITE_ACK_CMDFLAG;

        // check if header part is expected one
        if (memcmp(&cmpdt, &rcvdt.header, sizeof(cmpdt)) != 0) {
            result = -3;   // not a response to the request
        }

        // convert command(address) from network byte order to host byte order
        rcvdt.header.address = (unsigned int)IPAddress::NetworkToHostOrder( (int)rcvdt.header.address );
    }
    else { // short of data size
        result = -2;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_16bit
//  Function  : command communication(write request for 16bit value)
//  Arguments :
//      in comRef   : SiTCPcom class instance for referring UdpClient
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_16bit(SiTCPcom ^comRef, unsigned int cmd_addr, unsigned short wdata)
{
    int result;
    unsigned short value;

    value = (unsigned short)IPAddress::HostToNetworkOrder( (short)wdata );
    result = sitcp_write(comRef->cmdClient, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_32bit
//  Function  : command communication(write request for 32bit value)
//  Arguments :
//      in comRef   : SiTCPcom class instance for referring UdpClient
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_32bit(SiTCPcom ^comRef, unsigned int cmd_addr, unsigned int wdata)
{
    int result;
    unsigned int value;

    value = (unsigned int)IPAddress::HostToNetworkOrder( (int)wdata );
    result = sitcp_write(comRef->cmdClient, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_48bit
//  Function  : command communication(write request for 48bit value)
//  Arguments :
//      in comRef   : SiTCPcom class instance for referring UdpClient
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_48bit(SiTCPcom ^comRef, unsigned int cmd_addr, unsigned qword wdata)
{
    int result;
    unsigned char *p, *q;
    unsigned qword value, wvalue;

    wvalue = (unsigned qword)IPAddress::HostToNetworkOrder( (qword)wdata );

    value  = 0;
    p = (unsigned char*)(&value);
    q = (unsigned char*)(&wvalue);
    memcpy(&p[0], &q[2], 6);   // copy valid 48bit to upper bit of destination area

    result = sitcp_write(comRef->cmdClient, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_write_64bit
//  Function  : command communication(write request for 64bit value)
//  Arguments :
//      in comRef   : SiTCPcom class instance for referring UdpClient
//      in cmd_addr : command address
//      in wdata    : data to be written
//  Return : same as sitcp_write
//  Note : This function converts target data from host byte order
//         to network byte order
//#########################################################################
int sitcp_write_64bit(SiTCPcom ^comRef, unsigned int cmd_addr, unsigned qword wdata)
{
    int result;
    unsigned qword value;

    value  = (unsigned qword)IPAddress::HostToNetworkOrder( (qword)wdata );
    result = sitcp_write(comRef->cmdClient, cmd_addr, &value, sizeof(value));

    return(result);
}


//#########################################################################
//  Name      : sitcp_recv_cmd
//  Function  : command communication(receive response)
//  Arguments :
//      in cmdClient : instance of UdpClient class for command communication
//      in rbuffer   : buffer for received data
//      in timeout   : receive timeout value(msec)
//  Return : 
//      positive value: received data size
//      -11 : no response within a period of timeout value
//      -12 : no data
//  Note :
//#########################################################################
int sitcp_recv_cmd(UdpClient ^cmdClient, char *rbuffer, int timeout)
{
    int result = 0;

    // create an IPEndPoint instance that represents a remote host from which the data was sent
    IPEndPoint ^remoteEP = gcnew IPEndPoint( IPAddress::Parse(IPADDR_DEVICE),
                                             DEVICE_PORT_CTRL_DEVICE );

    // set timeout value
    cmdClient->Client->ReceiveTimeout = timeout;

    try {
        // wait until received data is ready
        array<Byte>^receiveBytes = cmdClient->Receive(remoteEP);
        if (receiveBytes->Length != 0) {
            pin_ptr<Byte> toRbuffer = &receiveBytes[0];
            memcpy(rbuffer, toRbuffer, receiveBytes->Length);
            result = receiveBytes->Length;
        } else {
            result = -12;
        }
    }
    catch(SocketException ^e) {
        // presume timeout;  e->ErrorCode = 10060(WSAETIMEDOUT)
        result = -11;
    }

    return(result);
}


//#########################################################################
//  Name      : sitcp_recv_data
//  Function  : receive data by fast data transfer(TCP)
//  Arguments :
//      in comRef  : instance of SiTCPcom class
//      in rbuffer : buffer for received data
//      in size    : size of data to be received(byte)
//      in timeout : receive timeout value(msec)
//  Return : 
//      positive value: received data size
//      -1 : communication error
//  Note :
//#########################################################################
#define TMP_BUF_SIZE (2048)
int sitcp_recv_data(SiTCPcom ^comRef, char *rbuffer, int size, int timeout)
{
    int tval, tval2nd;
    int result;
    int rbyte, rsize, req_size;
    unsigned qword count;

    result = 0;
    count = 0;

    // save timeout value for 1st read
    tval = timeout;
    // save timeout value for 2nd read or later
    tval2nd = timeout * 10;

    // create NetworkStream for receiving data
    NetworkStream^ ns = comRef->dataClient->GetStream();
    // create temporary buffer to store receive data 
    array<Byte>^ tmp_buff = gcnew array<Byte>(TMP_BUF_SIZE);

    do {
        if ((count <= 1) && (timeout > 0)) {
            // set timeout value
            ns->ReadTimeout = tval;

            // update timeout value for 2nd or later read
            tval = tval2nd;
        }

        // update request data size which fits temporary buffer
        rsize = size - result;
        req_size = (rsize < TMP_BUF_SIZE ) ? rsize : TMP_BUF_SIZE;

        try {
            // receive data
            rbyte = ns->Read(tmp_buff, 0, req_size);
        }
        catch (IO::IOException ^e) {
            // presume timeout
            break;
        }

        if (rbyte > 0) {
            // store received data to specified buffer
            pin_ptr<Byte> toRbuffer = &tmp_buff[0];
            memcpy(&rbuffer[result], toRbuffer, rbyte);

            // update total received data size
            result += rbyte;

            if (result >= size) {
                // finish when total received data size reached to request one
                break;
            }
        }
        else { // no data
            result = -1;
            break;
        }

        count++;
    } while(ns->DataAvailable);

    return(result);
}
